; Copyright (C) 2016 Texas Instruments Incorporated - http://www.ti.com/
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions
; are met:
;
; Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following disclaimer.
;
; notice, this list of conditions and the following disclaimer in the
; documentation and/or other materials provided with the
; distribution.
;
; Neither the name of Texas Instruments Incorporated nor the names of
; its contributors may be used to endorse or promote products derived
; from this software without specific prior written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;
; 	file:   cnc_app.asm
;
; 	brief:  PRU application to control 4 stepper motors over 4 us Ethernet
; 			includes sin^2 ramp generation for variable speeds.
;
;	Version		Description								Author
; 	0.1     	Created								  Thomas Leyrer

;include section
    .include "../include/macros/app_cnc_macros.inc"
	.include "../include/macros/macros.inc"
	.include "../include/macros/MACdef.inc"
	.include "../include/icss/icss_regs.inc"
	.include "../include/icss/icss_macros.inc"
	.include "../include/protocol/sorte_pru_master_register.inc"
	.include "../include/protocol/sorte_master_cfg.inc"

	.global CNC_APP_INIT
	.global	CNC_APP_EXE
	.global	CNC_APP_RET
	.global IDLE_WAIT_FOR_LINK_ACTIVE

; ------------------------------------------------------------------------------
; FUNCTION: CNC_APP_INIT
; ARGS:
; Description: setup table sin^2 for acceleration/de-acceleration and sound table
; ------------------------------------------------------------------------------

CNC_APP_INIT:
    mov         r0,r29
	M_INIT_SIN2_TBL
	M_INIT_AXEL_F_TBL
	ldi			TEMP_REG_1, AXEL_F_TBL_OFFSET
	ldi			TEMP_REG_2, 0
	sbco		&TEMP_REG_2,PRU0_DMEM_CONST, TEMP_REG_1, 4
	ldi32		TEMP_REG_1, STEP_TIME_50k_CONST
	sbco		&TEMP_REG_1,PRU0_DMEM_CONST, APPL_SPEED_OFFSET, 4
    ldi32		r21, 51200-64
	sbco		&r21, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET+4,4
	mov         r29,r0
	RET

; ------------------------------------------------------------------------------
; FUNCTION: CNC_APP_EXE
;
; Description: application to drive 4 stepper motors over 4us Ethernet
;
; ARGS:  	two modes of operation are supported which is position and dance
;			in position mode the motors spin the number steps programmed in
;			separate control field. The max speed of the movement is given in
;           another control field. In dance mode, a sequence of movements is
;			executed which result in a well known melody.
;  			APPL_BUTTON_OFFSET (0x00C0): byte 0 -> mode butten
;										 byte 3 -> start button
;			APPL_FADER_OFFSET  (0x00C4): number of steps for next movement
; 			APPL_SPEED_OFFSET  (0x00C8): speed is 1/f [MHz]*2^16
;
; ------------------------------------------------------------------------------
CNC_APP_EXE:
; ########## CNC APLLICATION ###########
; r24.b0 is state
; r24.b1 is substate
; r24.b2 as table index
; r24.b3 is state increment (0) or decrement (1) table index

; Configure MAC in MPY mode
; set mac mode to multiply only
	ldi		r25, 0
    xout 	0, &r25, 1

; check on host settings for app_cnc
    lbco 	&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,4
    qbeq    APP_CNC_SELECT, TEMP_REG_1.b0, MODE_BUTTON_DANCE
    qbne    APP_CNC_SKIP_POS_MODE, TEMP_REG_1.b0, MODE_BUTTON_POSITION

APP_CNC_SELECT:
; merge mode button position and dance using variables for
; 1/fstep
; step_time
; pause_time
; use register for these settings
; APPL_FSTEP_OFFSET			.set    0x00E0
; APPL_FSTEP_TIME_OFFSET	.set    0x00E4
; APPL_PAUSE_OFFSET			.set    0x00E8

; for dance mode load this three register from table
; table index:  AXEL_F_TBL_OFFSET
; table: AXEL_F_TBL_OFFSET+4

; read index first
    ldi			TEMP_REG_1.w0, AXEL_F_TBL_OFFSET
    lbco		&TEMP_REG_2.w0, PRU0_DMEM_CONST, TEMP_REG_1.w0, 2
; read table entry
	add			TEMP_REG_1.w0, TEMP_REG_1.w0,4
	add			TEMP_REG_1.w0, TEMP_REG_1.w0, TEMP_REG_2.w0
	lbco		&TEMP_REG_1, PRU0_DMEM_CONST, TEMP_REG_1.w0, 12

;	ldi32		TEMP_REG_2, note_A
	sbco		&TEMP_REG_1, PRU0_DMEM_CONST, APPL_FSTEP_OFFSET, 12
;	ldi32		TEMP_REG_2, 16*p125_A
;	sbco		&TEMP_REG_2, PRU0_DMEM_CONST, APPL_FSTEP_TIME_OFFSET, 4
;	ldi32		TEMP_REG_2, 8*p125
;	sbco		&TEMP_REG_2, PRU0_DMEM_CONST, APPL_PAUSE_OFFSET, 4

; get start button settings ready for next cycle.
    lbco 		&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,4
    mov			r22.b0,  TEMP_REG_1.b3

; power-off when no motor is selected
   	lbco 		&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,4
    qbne		APP_CNC_SKIP_IDLE_ALL, TEMP_REG_1.b3 ,0
    qbne 		APP_CNC_SKIP_IDLE_ALL, r24.b0, POSITION_WAIT1S
    ldi			TEMP_REG_1.w0, 0x7fff
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET
    sbco		&TEMP_REG_1.w0, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+4
    sbco		&TEMP_REG_1.w0, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+8
    sbco		&TEMP_REG_1.w0, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+12
    sbco		&TEMP_REG_1.w0, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    qba			APP_CNC_SKIP_POS_MODE


APP_CNC_SKIP_IDLE_ALL:

; execute postion mode - fader and ustep settings have no effect
; ********* state idle  *********

    qbne 	APP_CNC_SKIP_IDLE, r24.b0, POSITION_IDLE
	and	   	IO_CYCLE_CNT, IO_CYCLE_CNT, 0
	zero	&r22.w2, 10

; get a pulse on first step
	ldi		r22.w2, 5

; upate motor selection, r22.b1 first four bits are mapped to motors.
    mov		r22.b1, r22.b0

	ldi 	r24.b0, POSITION_WAIT1S
    ldi		r24.b1, ACCEL_STATE_INIT
    ldi		r24.b3, 0
;   set current values in out data
;   r23    = 4us_cnt
;   r22.w2 = sync_time

; reset IO
; in idle, clear output data and wait for 1 second using cycle counter
    zero	&r2,16
    ldi		TEMP_REG_2.w0, OUT_DATA_OFFSET
    sbco	&r2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 16

APP_CNC_SKIP_IDLE:
; ********* state wait 1(x) second  *********
; wait for 1 s using IO-CYCLE count which was cleared in idle state
    qbne 	APP_CNC_SKIP_WAIT, r24.b0, POSITION_WAIT1S
;    LDI32	TEMP_REG_1, 250000
    lbco	&TEMP_REG_1, PRU0_DMEM_CONST, APPL_PAUSE_OFFSET, 4
    qbgt	APP_CNC_SKIP_POS_MODE, IO_CYCLE_CNT, TEMP_REG_1
    and	   	IO_CYCLE_CNT, IO_CYCLE_CNT, 0
    lbco 	&r21, PRU0_DMEM_CONST, APPL_SPEED_OFFSET,4
    sbco 	&r21, PRU0_DMEM_CONST, APPL_SPEED_OFFSET+4, 4

   	lbco 	&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,4
    sbco 	&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET+16, 4

;   setup step size
;    ldi32	r21, 102400-64
    lbco	&r21, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET+4,4
    lbco 	&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,4
    qbne    APP_CNC_SKIP_IDLE_1, TEMP_REG_1.b0, MODE_BUTTON_DANCE
	lbco	&r21, PRU0_DMEM_CONST, APPL_FSTEP_TIME_OFFSET, 4

APP_CNC_SKIP_IDLE_1:
	ldi 	r24.b0, POSITION_QUARTER_TURN

; clear index into ramp curve
	ldi		r24.b2, 0

APP_CNC_SKIP_WAIT:

; ********* state quarter turn  *********
; cycle count is at 1
; step through rise, const, fall states of quarter turn
; use r24.b2 as table index
	qbne	APP_CNC_POS_QT_ACCEL_WAIT_INIT, r24.b1, ACCEL_STATE_INIT

; get step from LUT, remember first two entries are 32 bit and remaining 30 entries are 16 bit
    ldi		TEMP_REG_1.w0, SIN2_TBL32_OFFSET
    qble	APP_CNC_POS_QT_ACCEL_L1, r24.b2, 2
    lsl		TEMP_REG_2.b0, r24.b2, 2
    add		TEMP_REG_1.w0, TEMP_REG_1.w0, TEMP_REG_2.b0
    lbco	&sMAC.op1_32,PRU0_DMEM_CONST, TEMP_REG_1.w0, 4
    qba     APP_CNC_POS_QT_ACCEL_L2

APP_CNC_POS_QT_ACCEL_L1:
    lsl		TEMP_REG_2.b0, r24.b2, 1
    add     TEMP_REG_2.b0, TEMP_REG_2.b0, 4
    add		TEMP_REG_1.w0, TEMP_REG_1.w0, TEMP_REG_2.b0
    and     sMAC.op1_32, sMAC.op1_32, 0
    lbco	&sMAC.op1_32,PRU0_DMEM_CONST, TEMP_REG_1.w0, 2

APP_CNC_POS_QT_ACCEL_L2:

;   get time to wait on step by multiplying sin^2 with step time const
    lbco 	&sMAC.op2_32, PRU0_DMEM_CONST, APPL_SPEED_OFFSET+4, 4
;     ldi32   sMAC.op2_32, 0X00070000
;   in dance mode use different register
   	lbco 	&TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,4
    qbne    APP_CNC_POS_QT_ACCEL_L21, TEMP_REG_1.b0, MODE_BUTTON_DANCE
    lbco 	&sMAC.op2_32, PRU0_DMEM_CONST, APPL_FSTEP_OFFSET, 4

APP_CNC_POS_QT_ACCEL_L21:
	nop
    xin			0, &r26, 8

; result t_step [ns] is in R27.w0 (high word and R26.w2) low word -> r28
    mov 		r28.w0, r26.w2
    mov 		r28.w2, r27.w0
;    mov			r23, r28        ; keep in r23 for now

; multiply with 1/4000 * 2^32
    ldi32		sMAC.op2_32, FOUR_US_P32_CONST
	nop
    xin			0, &r26, 8

; result 4us_cnt is in r27
; now multiply 4us_cnt with 4000 and subtract from t_step
    mov			TEMP_REG_1, r27
    mov			r29, r27
    mov        	r23, r28
    ldi32		r28, 4000
    nop
    xin			0, &r26, 8

    sub 		r29, r23, r26

; restore 4us_cnt into r28
    mov			r28, TEMP_REG_1
; sync_time (0..3999 ns) is in r29

; caluclate final data consider offset from last step
; add prev sync_time to currennt sync time
    add			r29, r29, r22.w2
    ldi			TEMP_REG_1.w0, 4000
    qbgt		APP_CNC_POS_QT_ACCEL_L3, r29.w0, TEMP_REG_1.w0
    add			r28, r28, 1
    sub			r29.w0, r29, TEMP_REG_1.w0

APP_CNC_POS_QT_ACCEL_L3:
;   set current values in out data
;   update sync_time boundaries to 5..3990 ns
	ldi			TEMP_REG_1.w0, 3990
    qbgt		APP_CNC_POS_QT_ACCEL_L4, r29.w0, TEMP_REG_1.w0
    ldi			r29.w0, 3990

APP_CNC_POS_QT_ACCEL_L4:
    qblt		APP_CNC_POS_QT_ACCEL_L5, r29.w0, 5
    ldi			r29.w0, 5

APP_CNC_POS_QT_ACCEL_L5:
;   r23    = 4us_cnt
;   r22.w2 = sync_time
    lbco 	    &TEMP_REG_1, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET+16, 4
    qbbc		APP_CNC_SKIP_OUT_1, TEMP_REG_1.b3 ,0
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET
    sbco		&r22.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
APP_CNC_SKIP_OUT_1:
    qbbc		APP_CNC_SKIP_OUT_2, TEMP_REG_1.b3 ,1
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+4
    sbco		&r22.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2

APP_CNC_SKIP_OUT_2:
    qbbc		APP_CNC_SKIP_OUT_3, TEMP_REG_1.b3 ,2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+8
    sbco		&r22.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2

APP_CNC_SKIP_OUT_3:
    qbbc		APP_CNC_SKIP_OUT_4, TEMP_REG_1.b3 ,3
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+12
    sbco		&r22.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
APP_CNC_SKIP_OUT_4:

SKIP_DBG_CNC_1
    mov			r22.w2, r29.w0
    mov     	r23, r28

; move to state wait
    ldi 		r24.b1, ACCEL_STATE_WAIT_INIT
; no wait when r28 is 1 which means next cyle
    qbge		APP_CNC_POS_QT_ACCEL_L6, r28, 1
    sub			r28,r28,1
    qba			APP_CNC_SKIP_POS_MODE

;APP_CNC_POS_QT_ACCEL_L6:
; special case where r28 = 0
;    halt

APP_CNC_POS_QT_ACCEL_WAIT_INIT:

	qbne		APP_CNC_POS_QT_ACCEL_WAIT, r24.b1, ACCEL_STATE_WAIT_INIT
; ACCEL_STATE_WAIT_INIT
; clear sync_time during wait
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET
    ldi			TEMP_REG_2.w2, 0
    sbco		&TEMP_REG_2.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+4
    sbco		&TEMP_REG_2.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+8
    sbco		&TEMP_REG_2.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			TEMP_REG_2.w0, OUT_DATA_OFFSET+12
    sbco		&TEMP_REG_2.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2

APP_CNC_POS_QT_ACCEL_L6:
; increment ramp table index
    qbne        APP_CNC_POS_QT_DEACCEL, r24.b3,0
	add			r24.b2, r24.b2, 1
    qbne        APP_CNC_POS_QT_ACCEL_WAIT_INIT_L1, r24.b2,32

; stay on last step entry for 512000-64 steps
    ldi			r24.b2, 31
    sub			r21,r21,1
    qbne		APP_CNC_POS_QT_ACCEL_WAIT_INIT_L1, r21, 0

; walk table backwards,i.e deaccel
    ldi			r24.b3,1

APP_CNC_POS_QT_DEACCEL:
	sub			r24.b2, r24.b2, 1
	sub			TEMP_REG_1.b0, r24.b2 ,1
    qbne        APP_CNC_POS_QT_ACCEL_WAIT_INIT_L1, TEMP_REG_1.b0, 0
    ldi			r24.w0, 0
    ldi			TEMP_REG_2.w2, 0
    sbco		&TEMP_REG_2.w2, PRU0_DMEM_CONST, TEMP_REG_2.w0, 2
    ldi			r22.w2, 0
; done with this turn / tone
; jump to new table index for dance
    ; update index first
    lbco 		&TEMP_REG_1.b0, PRU0_DMEM_CONST, APPL_BUTTON_OFFSET,1
    qbne    	APP_CNC_POS_QT_DEACCEL_L1, TEMP_REG_1.b0, MODE_BUTTON_DANCE
    ldi			TEMP_REG_1.w0, AXEL_F_TBL_OFFSET
    lbco		&TEMP_REG_2.w0, PRU0_DMEM_CONST, TEMP_REG_1.w0, 2
    add			TEMP_REG_2.w0, TEMP_REG_2.w0, 12
    ldi			TEMP_REG_1.w2, 23*12
    qbne		APP_CNC_POS_QT_DEACCEL_L1,TEMP_REG_2.w0, TEMP_REG_1.w2  ; 23 tones
    ldi			TEMP_REG_2.w0, 0

APP_CNC_POS_QT_DEACCEL_L1:
    sbco		&TEMP_REG_2.w0, PRU0_DMEM_CONST, TEMP_REG_1.w0, 2
    qba			APP_CNC_SKIP_POS_MODE

APP_CNC_POS_QT_ACCEL_WAIT_INIT_L1:
    ldi 		r24.b1, ACCEL_STATE_WAIT

APP_CNC_POS_QT_ACCEL_WAIT:
	qbne		APP_CNC_POS_QT_ACCEL, r24.b1, ACCEL_STATE_WAIT
	qbeq		APP_CNC_POS_QT_ACCEL_WAIT_L2, r28, 0
    sub     	r28, r28,1
APP_CNC_POS_QT_ACCEL_WAIT_L2:
    qbne		APP_CNC_POS_QT_ACCEL_WAIT_L1, r28, 0
; jump back to new step pulse
    ldi			r24.b1, ACCEL_STATE_INIT
APP_CNC_POS_QT_ACCEL_WAIT_L1:

APP_CNC_POS_QT_ACCEL:

APP_CNC_SKIP_POS_MODE:

; check non next mode
; load operands

	jmp			CNC_APP_RET
